#include <gmsh.h>
#include <iostream>
#include <map>
#include <sstream>
#include <Eigen/Dense>
#include <cmath>
#include <iomanip>
#include <stdio.h>
#include <chrono>
#include "functionsBEM.hpp"
#include <omp.h>
#include <algorithm>
#include <list>

/*
 *  Verification
 */
void dispNodeCoordFun(const std::vector<size_t> &allNodeTags,
                      const std::vector<double> &allCoord)
{
    for(size_t i = 0; i < allNodeTags.size(); i++)
    {
        std::cout << "Node '" << allNodeTags[i] << "': (" << allCoord[3*i] << "; " << allCoord[3*i + 1] << "; " << allCoord[3*i + 2] << ")\n";
    }
    std::cout << std::endl;
}

void displayTime(const std::chrono::time_point<std::chrono::system_clock> start,
                  const std::chrono::time_point<std::chrono::system_clock> end,
                  std::string function_name, bool &showTime)
{
    //displays the time of the desired code section
    if(showTime)
    {
        std::cout << std::endl;
        std::cout << function_name << std::endl;
        std::cout << "Elapsed time in nanoseconds: "
        << std::chrono::duration_cast<std::chrono::nanoseconds>(end - start).count()
        << " ns" << std::endl;

        std::cout << "Elapsed time in microseconds: "
        << std::chrono::duration_cast<std::chrono::microseconds>(end - start).count()
        << " µs" << std::endl;

        std::cout << "Elapsed time in milliseconds: "
        << std::chrono::duration_cast<std::chrono::milliseconds>(end - start).count()
        << " ms" << std::endl;

        std::cout << "Elapsed time in seconds: "
        << std::chrono::duration_cast<std::chrono::seconds>(end - start).count()
        << " sec";
        std::cout << std::endl;
        std::cout << std::endl;
    }
}

void convergenceFun(const std::vector<int> &domainEntityTags,
                    const std::vector<std::vector<int>> &domainElementTypes,
                    const std::vector<std::vector<std::vector<size_t>>> &domainElementTags,
                    const std::vector<std::vector<std::vector<std::vector<double>>>> &elementalPhi,
                    const std::vector<std::vector<std::vector<std::vector<double>>>> &elementalElectricField)
{
    double l = 1.0;
    double phiLeft = 100.0;
    double phiRight = 200.0;
    double exactE = (phiLeft - phiRight) / l;

    double xLeft = 0.25;
    double xRight = 0.75;
    double yBottom = 0.25;
    double yTop = 0.75;

    double errorPhi = 0.0;
    double errorE = 0.0;
    double errorEWindow = 0.0;
    int count = 0;
    int countWindow = 0;

    std::vector<double> barycenters;

    for(std::size_t i = 0; i < domainEntityTags.size(); i++)
    {
        for(std::size_t j = 0; j < domainElementTypes[i].size(); j++)
        {
            gmsh::model::mesh::getBarycenters(domainElementTypes[i][j], domainEntityTags[i], false, false, barycenters);
            for(std::size_t k = 0; k < domainElementTags[i][j].size(); k++)
            {
                double numericalPhi = elementalPhi[i][j][k][0];
                double numericalE = elementalElectricField[i][j][k][0];

                double x = barycenters[3*k];
                double y = barycenters[3*k + 1];

                double exactPhi = phiLeft + (phiRight - phiLeft) * x / l;

                double relErrorPhi = abs( (exactPhi - numericalPhi) / exactPhi );
                double relErrorE = abs( (exactE - numericalE) / exactE);

                if(x > xLeft && x < xRight && y > yBottom && y < yTop)
                {
                    errorEWindow += relErrorE*relErrorE;
                    countWindow++;
                }

                errorPhi += relErrorPhi*relErrorPhi;
                errorE += relErrorE*relErrorE;
                count++;
            }
        }
    }
    errorPhi = sqrt(errorPhi / count);
    errorE = sqrt(errorE / count);
    errorEWindow = sqrt(errorEWindow / countWindow);
    std::cout << "Relative error on the electric potential: " << errorPhi << "\n";
    std::cout << "Relative error on the electric field: " << errorE << "\n";
    std::cout << "Relative error on the electric field (without the edges of the domain): " << errorEWindow << "\n\n";
}